static char *sccsid = "%M% %I% %D%";
/**********************************************************************
 *  DCOPYU   UNIX portion of DOS to UNIX file copy system.
 *
 *	(c) Copyright GMS, 1984,89.
 *	(c) DigiBoard Inc. 1989.
 *
 *	Based on the KERMIT Protocol.
 *
 *	UNIX to DOS file transfer utility.  This program is invoked
 *	the DOS companion program DCOPY.EXE.  Together they form a
 *	a communications link that performs DOS like file COPYs,
 *	between DOS and UNIX.
 *
 */

#include <stdio.h>
#include <ctype.h>
#include <signal.h>
#include <setjmp.h>
#include <fcntl.h>
#include <termio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/errno.h>

#define TIMEOUT 
#define FLUSH

/* Symbol Definitions */

#define XMITPACKSIZ  	94				/* packet size */
#define MAXPACKSIZ  	1024			/* Maximum packet size */
#define SOH				1				/* Start of header */
#define CR				13				/* ASCII Carriage Return */
#define SP				32				/* ASCII space */
#define DEL				127			   	/* Delete (rubout) */
#define XON				17				/* ASCII XON */

#define MAXTRY		5			/* Times to retry a packet */
#define MYQUOTE		'#'			/* Quote character I will use */
#define MYPAD		0			/* Number of padding characters I will need */
#define MYPCHAR		0			/* Padding character I need (NULL) */

#define MYTIME		5			/* Seconds after which I should be timed out */
#define MAXTIM		20			/* Maximum timeout interval */
#define MINTIM		2			/* Minumum timeout interval */

#define TRUE		-1			/* Boolean constants */
#define FALSE		0

#define DEFIM		FALSE 		/* Default image mode */

#define MYEOL		CR			/* End-Of-Line character I need */
#define MYQBIN		'N'			/* Binary quote character, N = will not do */
#define MYCHKTYP	'1'			/* Check sum type I need */
#define MYREPCHR	' '			/* Repeat count prefix I need, 0=none,  '~' */
#define MYWINDOWS	0			/* Number of windows */
#define MYCAPMASK	0x02		/* Capability mask: 000010 */


/* Macro Definitions */

/**********************************************************************
 * tochar: converts a control character to a printable one by adding a space.
 *
 * unchar: undoes tochar.
 *
 * ctl:	converts between control characters and printable characters by
 *		toggling the control bit (ie. ^A becomes A and A becomes ^A).
 *
 * unpar:  turns off the parity bit.
 */

#define tochar(ch)  ((ch) + ' ')
#define unchar(ch)  ((ch) - ' ')
#define ctl(ch)		((ch) ^ 64 )
#define unpar(ch)   ((ch) & 127)

/* Conditional Global Variables */

struct termio save_ttybuf;

/* Global Variables */

int	size,						/* Size of present data */
	rpsiz,						/* Maximum receive packet size */
	spsiz,						/* Maximum send packet size */
	longflg,					/* Long packet flag: 0=off n=on, set from rapr() */
	pad,						/* How much padding to send */
	timint,						/* Timeout for foreign host on sends */
	n,							/* Packet number */
	numtry,						/* Times this packet retried */
	oldtry,						/* Times previous packet retried */
	ttyfd,						/* File descriptor of tty for I/O */
	image,						/* -1 means 8-bit mode */
	autochk,					/* TRUE means auto-checking */
	debug = 0,					/* Indicates level of debugging output */
								/* 0 = none */
								/* 1 = prints out packets info */
								/* 2 = prints out 1 and more */
								/* 3 = prints out 1 and 2 and all packet data */
	baudrate,					/* baudrate used to set rpsiz */
	filecount,					/* Number of files left to send */
	use_pnam;

char
	state,						/* Present state of the automaton */
	cchksum,					/* Our (computed) checksum */
	padchar,					/* Padding character to send */
	eol,						/* End-Of-Line character to send */
	quote,						/* Quote character in incoming data */
	qbin,						/* Quote binary characters in incoming data */
	chktyp,						/* Check sum type */
	repchr,						/* Repeat prefix character in incoming data */
	capmask,					/* Capability mask */
	windows,					/* Number of windows */
	**filelist,					/* List of files to be sent */
	*filnam,					/* Current file name */
	recpkt[MAXPACKSIZ], 		/* Receive packet buffer */
	packet[MAXPACKSIZ],			/* Packet buffer */
	pathnam[512];

FILE	*fp;					/* File pointer for current disk file */

FILE	*logfile;				/* error log file pointer */

jmp_buf env;					/* Environment ptr for timeout longjump */

extern errno;

/**********************************************************************
 *  m a i n
 *
 *  Main routine - parse command and options, set up the
 *  tty lines, and dispatch to the appropriate routine.
 */

main(argc,argv)
int argc;								/* Character pointers to and count of */
char **argv;							/* command line arguments */
{
	char *cp;							/* char pointer */
	int rflg, sflg;						/* flags for CONNECT, RECEIVE, SEND */

	if (argc < 2) {       				/* Make sure there's a command line */
		printf("Args missing\n");
		usage();
		exit(-1);
		}

	cp = *++argv; argv++; argc -= 2;	/* Set up pointers to args */

	/*  Initialize these values and hope the first packet will get across OK */

	eol = CR;						/* EOL for outgoing packets */
	quote = '#';					/* Standard control-quote char "#" */
	pad = 0;						/* No padding */
	padchar = 0;					/* Use null if any padding wanted */

	sflg = rflg = 0;				/* Turn off all parse flags */

	image = DEFIM;					/* Default image mode */
	autochk = TRUE;

	if (*cp++ != '-') {
		printf("Bad Arg\n");
		usage();
		exit(-1);
		}

	while (*cp)
		switch (*cp++)
		{
		case 'B':					/* Block Size */
			rpsiz = atoi(cp);
			while (isdigit(*cp)) cp++;
			break;

		case 's':					/* S = Send command */
			sflg++;
			break;

		case 'r':					/* R = Receive command */
			rflg++;
			break;

		case 'd':					/* D = debug */
			if (isdigit(*cp))
				debug = *cp++ - '0';
			else debug = 1;
			break;
		
		case 'i':					/* I = Image (8-bit) mode */
			image = TRUE;
			break;

		case 't':					/* Disable auto type checking */
			autochk = FALSE;
			break;

		default:
			usage();
			exit(-1);
		}

	if ((sflg+rflg) != 1) {			/* Only one command allowed */
		usage();
		exit(-1);
		}

	ttopen("/dev/tty");				/* tty is the communications line) */
	ttbin();						/* Put the tty in binary mode */
	
	printf("DCOPYU 1.20 (c)DigiBoard 1986,89 .....\n");
	sleep(1);
	printf("DCOPYU 1.20 (c)DigiBoard 1986,89 .....\n");

	if (rpsiz <= 0)
	{
			rpsiz = 90;
			timint = MYTIME;
			if (baudrate <= B300) 
			{
				rpsiz = 60;
				timint = 10;
			} 
			else if (baudrate < B1200) 
			{
				rpsiz = 90;
			} 
			else if (baudrate < B4800) 
			{
				rpsiz = 240;
			} 
			else rpsiz = 480;
	}

	/* All set up, now execute the command that was given. */

	if (debug)
	{
		if ((logfile = fopen("dcopyu.log", "w")) == NULL)
		{
			printf("cannot open dcopyu.log file\n");
			Done(1);
		}
		LOG("Debugging level = %d\n\n",debug);
		if (sflg) LOG("Send command\n\n");
		if (rflg) LOG("Receive command\n\n");
	}
  
	if (debug) LOG("default baudrate = %d, rpsiz = %d\n", baudrate, rpsiz);

	if (sflg)								/* Send command */ 
	{
		if (argc--) filnam = *argv++;		/* Get file to send */
		else
		{
			usage();
			Done(0);
		}
		if (debug) LOG("sending %s\n", filnam);
		fp = NULL;							/* Indicate no file open yet */
		filelist = argv;					/* Set up the rest of the file list */
		filecount = argc;					/* Number of files left to send */
		if (sendsw() == FALSE)				/* Send the file(s) */
			LOG("Send failed.");			/* Report failure */
		else								/*  or */
			LOG("done.");					/* success */
	}

	else if (rflg)							/* Receive command */
	{
		if (argc--)
		{
			struct stat sbuf;

			strcpy(pathnam, *argv++);
			if (debug) LOG("optional pathname = %s\n", pathnam);

			use_pnam = 0;
			if (stat(pathnam, &sbuf) < 0)
			{
				if (errno == ENOENT) use_pnam = 1;
				else
				{
					sleep(1);
					error("Bad pathname: %s\n", pathnam);
					Done(0);
				}
			}
			else
			{
				if (sbuf.st_mode & S_IFDIR)
				{
					use_pnam = 2;
					strcat(pathnam, "/");
				}
				else use_pnam = 1;
			}
		}

		if (recsw() == FALSE)				/* Receive the file(s) */
			LOG("Receive failed.\n");
		else								/* Report failure */
			LOG("done.\n");					/* or success */
	}
	Done(0);
}

/********************************************************************
* trapexit - signal trap exit
*
*	This routine is is called by signal(2) when SIGINT or SIGQUIT are
*	trapped.  When called, it send an ERROR packet, resets the tty
*	and exits with a -9 return to UNIX.
*
*	The primary purpose of this routine is to prevent the exit back 
*	to UNIX without reseting the tty.
*
*	Added: 6/4/89, R.Larsen
*/
trapexit()
{
	error("Quit/Int Signal");			/* Send error packet to DCOPY */
	Done(-9);							/* Cleanup tty and exit(-9)	  */
	}
	
/*********************************************************
* Done - Restore tty and exit, return exit code s.
*/
Done(s)
int s;
{
	sleep(1);
	ioctl(ttyfd, TCFLSH, 2);
	ioctl(ttyfd, TCSETAF, &save_ttybuf);
	if(logfile) {
		fclose(logfile);
		chmod("dcopyu.log", 0666);
		}
	exit(s);
}

/**********************************************************************
 *  s e n d s w
 *
 *  Sendsw is the state table switcher for sending files.  It loops until
 *  either it finishes, or an error is encountered.  The routines called
 *  by sendsw are responsible for changing the state.
 *
 */

sendsw()
{
	char sinit(), sfile(), sdata(), seof(), sbreak();

	state = 'S';					/* Send initiate is the start state */
	n = 0;							/* Initialize message number */
	numtry = 0;						/* Say no tries yet */
	while(TRUE)						/* Do this as long as necessary */
	{
		if (debug) LOG("sendsw state: %c\n",state);
		switch(state)
		{
			case 'S': 				/* Send-Init */
				state = sinit();
				break;
			case 'F': 				/* Send-File */
				state = sfile();
				break;
			case 'D': 				/* Send-Data */
				state = sdata();
				break;
			case 'Z': 				/* Send-End-of-File */
				state = seof();
				break;
			case 'B': 				/* Send-Break */
				state = sbreak();
				break;
			case 'C':	 			/* Complete */
				return (TRUE);	
			case 'A':	 			/* Abort */
				return (FALSE);	
			default:	 			/* Unknown, fail */
				return (FALSE);	
		}
	}
}


/**********************************************************************
 *  s i n i t
 *
 *  Send Initiate: send this host's parameters and get other side's back.
 */

char sinit()
{
	int num, len;								/* Packet number, length */

	if (debug) LOG("In sinit retries: %d\n",numtry);

	if (numtry++ > MAXTRY) return('A'); 		/* If too many tries, give up */
	len = spar(packet);							/* Fill up init info packet */

#ifdef noFLUSH
	ioctl(ttyfd, TCFLSH, 0);					/* Flush pending input */
#endif

	spack('S',n,len,packet);					/* Send an S packet */
	switch(rpack(&len,&num,recpkt))				/* What was the reply? */
	{
	case 'N': 
		return(state);							/* NAK, try it again */

	case 'Y':									/* ACK */
		if (n != num)							/* If wrong ACK, stay in S state */
			return(state);						/* and try again */
		rpar(recpkt, len);						/* Get other side's init info */

		if (eol == 0) eol = '\n';				/* Check and set defaults */
		if (quote == 0) quote = '#';

		numtry = 0;								/* Reset try counter */
		n = (n+1)%64;							/* Bump packet count */
		return('F');							/* OK, switch state to F */

	case 'E':									/* Error packet received */
		prerrpkt(recpkt);						/* Print it out and */
		return('A');							/* abort */

	case FALSE:
		return(state);							/* Receive failure, try again */

	default: return('A');						/* Anything else, just "abort" */
	}
 }


/**********************************************************************
 *  s f i l e
 *
 *  Send File Header.
 */

char sfile()
{
	int num, len;							/* Packet number, length */
	char filnam1[50],						/* Converted file name */
	*newfilnam,								/* Pointer to file name to send */
	*cp;									/* char pointer */

	if (numtry++ > MAXTRY) return('A'); /* If too many tries, give up */
	
	if (fp == NULL)							/* If not already open, */
	{
		if (debug) LOG("   Opening %s for sending.\n",filnam);
		fp = fopen(filnam,"r");				/* open the file to be sent */
		if (fp == NULL)						/* If bad file pointer, give up */
		{
			error("File: '%s' not found.",filnam);
			return('A');
		}
	}

	strcpy(filnam1, filnam);				/* Copy file name */
	newfilnam = cp = filnam1;
	while (*cp)								/* Strip off all leading directory */
	if (*cp++ == '/')						/* names (ie. up to the last /). */
		newfilnam = cp;

	if (autochk)
	{
		if (isbinfile(fp)) strcat(newfilnam, ";B");
		else strcat(newfilnam, ";A");
		rewind(fp);
	}
	len = strlen(newfilnam);

	LOG("Sending %s as %s\n",filnam,newfilnam);

	spack('F',n,len,newfilnam);				/* Send an F packet */
	switch(rpack(&len,&num,recpkt))			/* What was the reply? */
	{			
	case 'N':								/* NAK, just stay in this state, */
		num = (--num<0 ? 63:num);			/* unless it's NAK for next packet */
		if (n != num)						/* which is just like an ACK for */ 
		return(state);						/* this packet so fall thru to... */

	case 'Y':								/* ACK */
		if (n != num) return(state); 		/* If wrong ACK, stay in F state */
		numtry = 0;							/* Reset try counter */
		n = (n+1)%64;						/* Bump packet count */
		if ((size = bufill(packet)) == EOF)
			return('Z');					/* Get first data from file */
		return('D');						/* Switch state to D */

	case 'E':								/* Error packet received */
		prerrpkt(recpkt);					/* Print it out and */
		return('A');						/* abort */

	case FALSE: return(state);				/* Receive failure, stay in F state */

	default:	return('A');				/* Something else, just "abort" */
	}
}


/**********************************************************************
 *  s d a t a
 *
 *  Send File Data
 */

char sdata()
{
	int num, len;

	if (numtry++ > MAXTRY) return('A'); 	/* If too many tries, give up */

	spack('D',n,size,packet);				/* Send a D packet */
	switch(rpack(&len,&num,recpkt))			/* What was the reply? */
	{			
	case 'N':								/* NAK, just stay in this state, */
		num = (--num<0 ? 63:num);			/* unless it's NAK for next packet */
		if (n != num)						/* which is just like an ACK for */
			return(state);					/* this packet so fall thru to... */
		
	case 'Y':								/* ACK */
		if (n != num) return(state); 		/* If wrong ACK, fail */
		numtry = 0;							/* Reset try counter */
		n = (n+1)%64;						/* Bump packet count */
		if ((size = bufill(packet)) == EOF) /* Get data from file */
			return('Z');					/* If EOF set state to that */
		return('D');						/* Got data, stay in state D */

	case 'E':								/* Error packet received */
		prerrpkt(recpkt);					/* Print it out and */
		return('A');						/* abort */

	case FALSE: return(state);				/* Receive failure, stay in D */

	default:	return('A');				/* Anything else, "abort" */
	}
}


/**********************************************************************
 *  s e o f
 *
 *  Send End-Of-File.
 */

char seof()
{
	int num, len;							/* Packet number, length */
	if (numtry++ > MAXTRY) return('A'); 	/* If too many tries, "abort" */

	spack('Z',n,0,packet);					/* Send a 'Z' packet */
	switch(rpack(&len,&num,recpkt))			/* What was the reply? */
	{
	case 'N':								/* NAK, just stay in this state, */
		num = (--num<0 ? 63:num);			/* unless it's NAK for next packet, */
		if (n != num)						/* which is just like an ACK for */
		return(state);						/* this packet so fall thru to... */

	case 'Y':			/* ACK */
		if (n != num) return(state); 		/* If wrong ACK, hold out */
		numtry = 0;							/* Reset try counter */
		n = (n+1)%64;						/* and bump packet count */
		if (debug) LOG("	  Closing input file %s, ",filnam);
		fclose(fp);							/* Close the input file */
		fp = NULL;							/* Set flag indicating no file open */ 

		if (debug) LOG("looking for next file...\n");
		if (gnxtfl() == FALSE)				/* No more files go? */
		return('B');						/* if not, break, EOT, all done */
		if (debug) LOG("	  New file is %s\n",filnam);
		return('F');						/* More files, switch state to F */

	case 'E':								/* Error packet received */
		prerrpkt(recpkt);					/* Print it out and */
		return('A');						/* abort */

	case FALSE: return(state);				/* Receive failure, stay in Z */

	default:	return('A');				/* Something else, "abort" */
	}
}


/**********************************************************************
 *  s b r e a k
 *
 *  Send Break (EOT)
 */

char sbreak()
{
	int num, len;						/* Packet number, length */
	if (numtry++ > MAXTRY) return('A'); /* If too many tries "abort" */

	spack('B',n,0,packet);				/* Send a B packet */
	switch (rpack(&len,&num,recpkt))	/* What was the reply? */
	{
	case 'N':							/* NAK, just stay in this state, */
		num = (--num<0 ? 63:num);		/* unless NAK for previous packet, */
		if (n != num)					/* which is just like an ACK for */
		return(state);					/* this packet so fall thru to... */

	case 'Y':							/* ACK */
		if (n != num) return(state); 	/* If wrong ACK, fail */
		numtry = 0;						/* Reset try counter */
		n = (n+1)%64;					/* and bump packet count */
		return('C');					/* Switch state to Complete */

	case 'E':							/* Error packet received */
		prerrpkt(recpkt);				/* Print it out and */
		return('A');					/* abort */

	case FALSE: return(state);	/* Receive failure, stay in B */

	default:	return ('A');	/* Other, "abort" */
   }
}


/**********************************************************************
 *  r e c s w
 *
 *  This is the state table switcher for receiving files.
 */

recsw()
{
	char rinit(), rfile(), rdata();		/* Use these procedures */

	state = 'R';						/* Receive-Init is the start state */
	n = 0;								/* Initialize message number */
	numtry = 0;							/* Say no tries yet */

	while(TRUE)
	{
		if (debug) LOG(" recsw state: %c\n",state);
		switch(state)					/* Do until done */
		{
			case 'R':	state = rinit(); break; /* Receive-Init */
			case 'F':	state = rfile(); break; /* Receive-File */
			case 'D':	state = rdata(); break; /* Receive-Data */
			case 'C':	return(TRUE);		/* Complete state */
			case 'A':	return(FALSE);		/* "Abort" state */
		}
	}
}

	
/**********************************************************************
 *  r i n i t
 *
 *  Receive Initialization
 */
  
char rinit()
{
	int len, num;								/* Packet length, number */

	if (numtry++ > MAXTRY) return('A'); 		/* If too many tries, "abort" */

	switch(rpack(&len,&num,packet))				/* Get a packet */
	{
	case 'S':									/* Send-Init */
		rpar(packet, len);						/* Get the other side's init data */
		len = spar(packet);						/* Fill up packet with my init info */
		spack('Y',n,len,packet);				/* ACK with my parameters */
		oldtry = numtry;						/* Save old try count */
		numtry = 0;								/* Start a new counter */
		n = (n+1)%64;							/* Bump packet number, mod 64 */
		return('F');							/* Enter File-Receive state */

	case 'E':									/* Error packet received */
		prerrpkt(recpkt);						/* Print it out and */
		return('A');							/* abort */

	case FALSE:									/* Didn't get packet */
		spack('N',n,0,0);						/* Return a NAK */
		return(state);							/* Keep trying */

	default:	 return('A');					/* Some other packet type, "abort" */
	}
}


/**********************************************************************
 *  r f i l e
 *
 *  Receive File Header
 */

char rfile()
{
	int num, len;							/* Packet number, length */
	char filnam1[200];						/* Holds the converted file name */
	char filnam2[200];

	if (numtry++ > MAXTRY) return('A'); 	/* "abort" if too many tries */

	switch(rpack(&len,&num,packet))			/* Get a packet */
	{
	case 'S':								/* Send-Init, maybe our ACK lost */
		if (oldtry++ > MAXTRY) return('A'); /* If too many tries "abort" */
		if (num == ((n==0) ? 63:n-1)) 		/* Previous packet, mod 64? */
		{									/* Yes, ACK it again with  */
			len = spar(packet);				/* our Send-Init parameters */
			spack('Y',num,len,packet);
			numtry = 0;						/* Reset try counter */
			return(state);					/* Stay in this state */
		}
		else return('A');					/* Not previous packet, "abort" */

	case 'Z':								/* End-Of-File */
		if (oldtry++ > MAXTRY) return('A');
		if (num == ((n==0) ? 63:n-1)) 		/* Previous packet, mod 64? */
		{									/* Yes, ACK it again. */
			spack('Y',num,0,0);
			numtry = 0;
			return(state);					/* Stay in this state */
		}
		else return('A');					/* Not previous packet, "abort" */

	case 'F':								/* File Header (just what we want) */
		if (num != n) return('A');			/* The packet number must be right */
		strcpy(filnam1, packet);			/* Copy the file name */

		if (autochk)
		{
			char *p;

			for (p=filnam1; *p && (*p != ';'); p++)
				;
			if (*p == ';')						/* get auto mode */
			{
				*p++ = '\0';					/* string terminate file name */
				if (*p == 'B') {
					if (debug) LOG("File is Binary\n");
					image = TRUE;
				}
				else if (*p == 'A') {
					if (debug) LOG("File is Text\n");
					image = FALSE;
				}
				else
				{
					error("Unknown File Type\n");
					return('A');
				}
			}
		}
		if (use_pnam == 1) strcpy(filnam1, pathnam);
		else if (use_pnam == 2)
		{
			char tmpfile[512];
			strcpy(tmpfile, pathnam);
			strcat(tmpfile, filnam1);
			strcpy(filnam1, tmpfile);
		}

		if (debug) LOG("filename = %s\n", filnam1);

		if ((fp=fopen(filnam1,"w"))==NULL) 			/* Try to open a new file */
		{
			error("Cannot create file: %s",filnam1); /* Give up if can't */
			return('A');
		}
		else								/* OK, give message */
			LOG("Receiving %s as %s\n",packet,filnam1);

		spack('Y',n,0,0);					/* Acknowledge the file header */
		oldtry = numtry;					/* Reset try counters */
		numtry = 0;							/* ... */
		n = (n+1)%64;						/* Bump packet number, mod 64 */
		return('D');						/* Switch to Data state */

	case 'B':								/* Break transmission (EOT) */
		if (num != n) return ('A'); 		/* Need right packet number here */
		spack('Y',n,0,0);					/* Say OK */
		return('C');						/* Go to complete state */

	case 'E':								/* Error packet received */
		prerrpkt(recpkt);					/* Print it out and */
		return('A');						/* abort */

	case FALSE:								/* Didn't get packet */
		spack('N',n,0,0);					/* Return a NAK */
		return(state);						/* Keep trying */

	default:	return ('A');				/* Some other packet, "abort" */
	}
}


/**********************************************************************
 *  r d a t a
 *
 *  Receive Data
 */

char rdata()
{
	int num, len;								/* Packet number, length */
	if (numtry++ > MAXTRY) return('A'); 		/* "abort" if too many tries */

	switch(rpack(&len,&num,packet))				/* Get packet */
	{
	case 'D':									/* Got Data packet */
		if (num != n)							/* Right packet? */
		{										/* No */
		if (oldtry++ > MAXTRY)
			return('A');						/* If too many tries, abort */
		if (num == ((n==0) ? 63:n-1)) 			/* Else check packet number */
		{										/* Previous packet again? */
			spack('Y',num,6,packet); 			/* Yes, re-ACK it */
			numtry = 0;							/* Reset try counter */
			return(state);						/* Don't write out data! */
		}
		else return('A');						/* sorry, wrong number */
		}
												/* Got data with right packet number */
		bufemp(packet,len);						/* Write the data to the file */
		spack('Y',n,0,0);						/* Acknowledge the packet */
		oldtry = numtry;						/* Reset the try counters */
		numtry = 0;								/* ... */
		n = (n+1)%64;							/* Bump packet number, mod 64 */
		return('D');							/* Remain in data state */

	case 'F':									/* Got a File Header */
		if (oldtry++ > MAXTRY)
		return('A');							/* If too many tries, "abort" */
		if (num == ((n==0) ? 63:n-1)) 			/* Else check packet number */
		{										/* It was the previous one */
		spack('Y',num,0,0);						/* ACK it again */
		numtry = 0;								/* Reset try counter */
		return(state);							/* Stay in Data state */
		}
		else return('A');						/* Not previous packet, "abort" */

	case 'Z':									/* End-Of-File */
		if (num != n) return('A');				/* Must have right packet number */
		spack('Y',n,0,0);						/* OK, ACK it. */
		fclose(fp);								/* Close the file */
		n = (n+1)%64;							/* Bump packet number */
		return('F');							/* Go back to Receive File state */

	case 'E':									/* Error packet received */
		prerrpkt(recpkt);						/* Print it out and */
		return('A');							/* abort */

	case FALSE:									/* Didn't get packet */
		spack('N',n,0,0);						/* Return a NAK */
		return(state);							/* Keep trying */

	default:	 return('A');					/* Some other packet, "abort" */
	}
}


clkint(s)								/* signal interrupt handler */
{
	LOG("*** TIMEOUT OCCURED ***\n");
	longjmp(env,TRUE);					/* Tell rpack to give up */
}


/**********************************************************************
 *  s p a c k
 *
 *  Send a Packet
 */

spack(type,num,len,data)
char type, *data;
int num, len;
{
	int i;
	char buffer[1024];							/* Checksum, packet buffer */
	register char hchksum, chksum;
	register char *bufp;

	if (debug)
	{
		if (data != NULL) data[len] = '\0';
		LOG("*** spack type:  %c\n",type);
		LOG("           num:  %d\n",num);
		LOG("           len:  %d\n",len);
		if (data != NULL)
			LOG("          data: \"%s\"\n",data);
	}

	if ((len < 0) || (len > MAXPACKSIZ))
	{
		LOG("***Send Pkt len: %d Error.\n", len);
		return('A');
	}
  
	bufp = buffer;
	for (i=1; i<=pad; i++)
		write(ttyfd,&padchar,1); 				/* Issue any padding */

	*bufp++ = SOH;								/* Packet marker, ASCII 1 (SOH) */
	if (len+3 <= 94)
	{
		*bufp++ = tochar(len+3);				/* Send the character count */
		chksum  = tochar(len+3);				/* Initialize the checksum */
		*bufp++ = tochar(num);					/* Packet number */
		chksum += tochar(num);
		*bufp++ = type;							/* Packet type */
		chksum += type;
	}
	else
	{
		*bufp++ = tochar(0);					/* Send the character count */
		chksum  = tochar(0);					/* Initialize the checksum */
		*bufp++ = tochar(num);					/* Packet number */
		chksum += tochar(num);
		*bufp++ = type;							/* Packet type */
		chksum += type;
		*bufp   = tochar((len+6)/95);/* Lenght byte 1 */
		chksum += *bufp++;
		*bufp 	= tochar((len+6)%95);/* lenght byte 2 */
		chksum += *bufp++;
		hchksum = (chksum+((chksum & 192)/64)) & 63;
		*bufp   = tochar(hchksum);				/* Header Check sum */
		chksum += *bufp++;
	}

	for (i=0; i<len; i++)
	{
			*bufp++ = data[i];
			chksum += data[i];
	}
	chksum = (((chksum&0300) >> 6)+chksum)&077; /* Compute final checksum */
	*bufp++ = tochar(chksum);
	*bufp++ = eol;								/* Extra-packet line terminator */
	if (debug>2)
	{
			*bufp = '\0';
			LOG("\nSpack:%s",buffer);
	}
	write(ttyfd,buffer,bufp-buffer);			/* Send the packet */
}

/**********************************************************************
 *  r p a c k
 *
 *  Read a Packet
 */

rpack(len,num,data)
int *len, *num;							/* Packet length, number */
char *data;								/* Packet data */
{
	int i, done;						/* Data character number, loop exit */
	char t;								/* Current input character */
	char type;							/* Packet type */
	char rchksum;						/* Checksum received from other host */
	char hchksum;						/* Header Checksum */
	extern clkint();

#ifdef TIMEOUT
	if (setjmp(env)) return FALSE;			/* Timed out, fail */
	signal(SIGALRM,clkint);					/* Setup the timeout */
	if ((timint > MAXTIM) || (timint < MINTIM)) timint = MYTIME;
	if (debug>2) LOG("timeout = %d\n", timint);
	alarm(timint);
#endif

	if (debug>1) LOG("\nRpack:");
	while (inchr() != SOH);						/* Wait for packet header */

	done = FALSE;								/* Got SOH, init loop */
	while (!done)								/* Loop to get a packet */
	{
		cchksum = 0;
		if ((t = cinchr()) == SOH) continue; 	/* Resynchronize if SOH */
		*len = unchar(t)-3;						/* Character count */

		if ((t = cinchr()) == SOH) continue; 	/* Resynchronize if SOH */
		*num = unchar(t);						/* Packet number */

		if ((t = cinchr()) == SOH) continue; 	/* Resynchronize if SOH */
		type = t;								/* Packet type */

		/* Handle long packets */

		if (*len == -3)
		{
			if ((t = cinchr()) == SOH) continue; 	/* Resynchronize if SOH */
			*len = unchar(t) * 95;					/* Lenght X1 */

			if ((t = cinchr()) == SOH) continue; 	/* Resynchronize if SOH */
			*len += unchar(t) - 6;					/* Lenght X2 */

			if ((t = inchr()) == SOH) continue; 	/* Resynchronize if SOH */
			hchksum = (cchksum + ((cchksum & 192)/64)) & 63;
			if (hchksum != unchar(t))
			{
				if (debug) LOG("*** Header Checksum Error ***\n");
				return(FALSE);
			}
			cchksum += t;
		}

		if (*len > rpsiz) LOG("*** Packet lenght too long ***\n");
		if (*len < 0) LOG("*** Packet lenght < 0 ***\n");

		/* Put len characters from the packet into the data buffer */

		for (i=0; i<*len; i++)
		{
			if ((t = cinchr()) == SOH) continue; /* Resynch if SOH */
			if (i <= rpsiz) data[i] = t;
		}

		data[*len] = 0;							/* Mark the end of the data */

		if ((t = inchr()) == SOH) continue; 	/* Resynchronize if SOH */
		rchksum = unchar(t);					/* Convert to numeric */
		done = TRUE;							/* Got checksum, done */
	}

#ifdef TIMEOUT
	alarm(0);								/* Disable the timer interrupt */
#endif

	if (debug)								/* Display incoming packet */
	{
		if (data != NULL)
			data[*len] = '\0';				/* Null-terminate data to print it */
		LOG("  rpack type: %c\n",type);
		LOG("         num:  %d\n",*num);
		LOG("         len:  %d\n",*len);
		if (data != NULL)
			LOG("        data: \"%s\"\n",data);
	}
											/* Fold in bits 7,8 to compute */
	cchksum = (((cchksum&0300) >> 6)+cchksum)&077; /* final checksum */

	if (cchksum != rchksum)
	{
		if (debug) LOG("*** Checksum Error ***\n");
		return(FALSE);
	}

	return(type);							/* All OK, return packet type */
}


/**********************************************************************
 * Get a parity adjusted character from the line, add it to the checksum
 * and return it.
 */

cinchr()
{
	char ch;

	ch = inchr();					/* Get a character */
	if (!image) ch &= 0177;
	cchksum += ch;					/* Add to the checksum */
	return(ch);
}


/**********************************************************************
 * Get a character from the line.  Do any necessary parity stripping
 * and return the character.  
 */

inchr()
{
	char ch;

	read(ttyfd,&ch,1);
	if (debug>5)						/* If debugging put a copy in file */
		LOG("%c:%03o|",ch,ch);
	return(ch);
}


static int
nextc(p)
register int *p;
{
	register int c, n;

	n = 0;
	if ((*p=getc(fp)) != EOF) {
		++n;
		if(repchr)
			while((c=getc(fp))!=EOF) {  	/* Handle repeated characters */
			if (c != *p) {
				ungetc(c, fp);
				break;
				}
			++n;
			if(n == 94) break;
			}
		}
	else return(0);

#ifdef KMT_IMAGE
	if (kmt_image == 4) {
		*p &= 0xff;
		if (*p <= 0x0D) *p += 0xA0;
		else if (*p >= 0xA0 && *p <= 0xAD) *p -= 0xA0;
			 else if (*p == 0xFF) *p = 0xB3;
		}
#endif
#ifdef FAST
	if (kmt_image == 3) *p = fastab1[*p & 0xff];
#endif
	return(n);
	}


/**********************************************************************
 *  b u f i l l
 *
 *  Get a bufferful of data from the file that's being sent.
 *  Only control-quoting is done; 8-bit & repeat count prefixes are
 *  not handled.
 */

bufill(buffer)
char buffer[];									/* Buffer */
{
	register int i;								/* Loop index */
	int t;										/* Char read from file */
	int t7;										/* 7-bit version of above */
	int c;
	register int n;

	i = 0;										/* Init data buffer pointer */
	n = nextc(&c);
	while (c != EOF)
	{
		while (n)
		{
			t = c;
			if (n > 2 && repchr)
			{
				buffer[i++] = repchr;
				buffer[i++] = tochar(n);
				n = 0;
			}
			else --n;
			if (image) t7 = t;
			else t7 = unpar(t);						/* Get low order 7 bits */

			if (t7 < SP || t7==DEL || t7==quote)	/* Does this char require */
			{										/* special handling? */
				if (t=='\n' && !image)
				{									/* Do LF->CRLF mapping if !image */
					buffer[i++] = quote;
					buffer[i++] = ctl('\r');
				}
				buffer[i++] = quote;				/* Quote the character */
				if (t7 != quote)
				{
					t = ctl(t);						/* and uncontrolify */
					t7 = ctl(t7);
				}
			}

			if (repchr && (t == repchr))
				buffer[i++] = quote;

			if (image) buffer[i++] = t;				/* Deposit the character itself */
			else buffer[i++] = t7;
		}

		if (i >= spsiz-10) return(i);				/* Check length */

		n = nextc(&c);

	}
	if (i==0) return(EOF);							/* Wind up here only on EOF */
	return(i);										/* Handle partial buffer */
}


/**********************************************************************
 *	b u f e m p
 *
 *  Put data from an incoming packet into a file.
 */

bufemp(buffer,len)
char  buffer[];								/* Buffer */
int   len;									/* Length */
{
	register int i;							/* Counter */
	char t;									/* Character holder */
	int t7;
	int n;

	for (i=0; i<len; i++)					/* Loop thru the data field */
	{
		n = 1;
		t = buffer[i];						/* Get character */
		if (repchr && t == repchr)
		{
			n = unchar(buffer[++i]);
			t = buffer[++i];
		}
		if (t == MYQUOTE)					/* Control quote? */
		{									/* Yes */
			t = buffer[++i];				/* Get the quoted character */
			if (image) t7 = t & 0177;
			else t7 = t;
			/*if ((t & 0177) != MYQUOTE)		/* Low order bits match quote char? */
			if ((t7 >= '?') && (t7 <= '_'))		/* Ctrl char in range ? */
				t = ctl(t);						/* No, uncontrollify it */
		}
		if (t==CR && !image)				/* Don't pass CR if in image mode */
			continue;

		while(n--) putc(t,fp);
	}
}


/**********************************************************************
 *  g n x t f l
 *
 *  Get next file in a file group
 */

gnxtfl()
{

	if (debug) LOG("	  gnxtfl: filelist = \"%s\"\n",*filelist);
	filnam = *(filelist++);
	if (filecount-- == 0) return FALSE; 			/* If no more, fail */
	else return TRUE;								/* else succeed */
}


/**********************************************************************
 *  s p a r
 *
 *  Fill the data array with my send-init parameters
 *
 */

spar(data)
char data[];
{
	if (rpsiz > 94) data[0] = tochar(94);
	else data[0] = tochar(rpsiz); 		/* Biggest packet I can receive */
	data[1] = tochar(MYTIME);			/* When I want to be timed out */
	data[2] = tochar(MYPAD);			/* How much padding I need */
	data[3] = ctl(MYPCHAR); 			/* Padding character I want */
	data[4] = tochar(MYEOL);			/* End-Of-Line character I want */
	data[5] = MYQUOTE;					/* Control-Quote character I send */

	data[6] = MYQBIN;					/* Binary quote I want */
	data[7] = MYCHKTYP;					/* Check sum type I need */
	data[8] = MYREPCHR;					/* Repeat count prefix I need */
	data[9] = tochar(MYCAPMASK);		/* Mask of my capabilities */
	data[10] = tochar(MYWINDOWS);		/* Number of windows */
	data[11] = tochar(rpsiz/95);		/* Long packet size */
	data[12] = tochar(rpsiz % 95);
	return(13);
}


/*  r p a r
 *
 *  Get the other host's send-init parameters
 *
 */

rpar(data,n)
char data[];
int n;									/* Size of init packet */
{
	spsiz = unchar(data[0]); 			/* Maximum send packet size */
	timint = unchar(data[1]);			/* When I should time out */
	pad = unchar(data[2]);				/* Number of pads to send */
	padchar = ctl(data[3]);				/* Padding character to send */
	eol = unchar(data[4]);				/* EOL character I must send */
	quote = data[5];					/* Incoming data quote character */

	qbin=chktyp=repchr=capmask=windows=0;	/* Setup for old init packets */

	if(n>8) {
		qbin = data[6];						/* Incoming data bin quote */
		chktyp = data[7];					/* Incoming check type */
		repchr = data[8];					/* Incoming repeat char */
		}
	if(n>10) {
		capmask = unchar(data[9]);			/* Incoming capabilities */
		windows = unchar(data[10]);			/* Number of windows */
		longflg = 0;
		}		
	if (capmask & 0x02) {					/* Check for long packets */
		spsiz =  unchar(data[11]) * 95;		/*  Long packet size */
		spsiz += unchar(data[12]);
		longflg = 1;
		}
	if (spsiz == 0) spsiz = 80;
	if (spsiz > MAXPACKSIZ) spsiz = MAXPACKSIZ;
	if (timint == 0 ) timint = MYTIME;
	if (eol == 0) eol = CR;
	if (qbin == 'Y') qbin = MYQBIN;
	if (qbin == 'N') qbin = 0;
	if (qbin == ' ') qbin = 0;
	if (chktyp == ' ') chktyp = '1';
	if (repchr == ' ') repchr = 0;

	if (debug) {
		LOG("*** Recv Init Parms ***\n");
		LOG("# of Parms:  %d\n",n);
		LOG("Send Size:   %d\n",spsiz);
		LOG("My Timeout:  %d\n",timint);
		LOG("Send Pads:   %d\n",pad);
		LOG("Pad Char:    %d\n",padchar);
		LOG("Send EOL:    %d\n",eol);
		LOG("Ctl QChr:    %d\n",quote);

		LOG("Bin QChr:    %d\n",qbin);
		LOG("Check Type:  %d\n",chktyp);
		LOG("Repeat Chr:  %d\n",repchr);
		LOG("Cap Mask:    0x%02x\n",capmask);
		LOG("Windows:     %d\n",windows);
		}
	return(n);
}


/**********************************************************************
 *  Dcopy printing routines:
 *
 *  usage - print command line options showing proper syntax
 *  error - like printmsg
 *  prerrpkt - print contents of error packet received from remote host
 */

/**********************************************************************
 *  u s a g e 
 *
 *  Print summary of usage info and quit
 */

usage()
{

	error(" * Illegal arguments * ");
}

/**********************************************************************
 *  e r r o r
 *
 *  log error message.
 *
 *  send an error packet with the message.
 */

/*VARARGS1*/
error(fmt, a1, a2, a3, a4, a5)
char *fmt;
{
	char msg[80];
	int len;

	LOG(fmt, a1, a2, a3, a4, a5);
	sprintf(msg,fmt,a1,a2,a3,a4,a5); 		/* Make it a string */
	len = strlen(msg);
	spack('E',n,len,msg);					/* Send the error packet */
}

/**********************************************************************
 *  p r e r r p k t
 *
 *  Print contents of error packet received from remote host.
 */
prerrpkt(msg)
char *msg;
{
	LOG("error packet: %s\n",msg);
}

ttopen(tname)
char *tname;
{
	if ((ttyfd = open(tname, O_RDWR)) < 0)
	{
		LOG("cannot open %s\n", tname);
		Done(1);
	}
}

ttbin()
{
	struct termio ttybuf;

	ioctl(ttyfd, TCGETA, &ttybuf);
	save_ttybuf = ttybuf;

	/* Code added to trap signals, RPL */
	/* OK to arm trap since save_ttybuf is now valid */


	signal(SIGINT, trapexit);
	signal(SIGQUIT, trapexit);

	ttybuf.c_iflag |= (IXON|IXOFF);
/*
	ttybuf.c_iflag &= ~(IXON|IXOFF);
*/

	ttybuf.c_lflag &= ~(ICANON|ECHO);
	ttybuf.c_lflag |= ISIG;		/* do check for interrupt */
	ttybuf.c_iflag |= (BRKINT|IGNPAR);
	ttybuf.c_iflag &= ~(IGNBRK|INLCR|IGNCR|ICRNL|IUCLC|INPCK|ISTRIP|IXANY);
	ttybuf.c_oflag &= ~OPOST;
	ttybuf.c_cflag &= ~(CSIZE|PARENB);
	ttybuf.c_cflag |= (CS8|CREAD);
	ttybuf.c_cc[4] = 1;   /* [VMIN]  Maybe should be bigger for all Sys V? */
	ttybuf.c_cc[5] = 0;   /* [VTIME] Should be set high enough to ignore */
	baudrate = ttybuf.c_cflag & 0x0F;
	ioctl(ttyfd, TCSETA, &ttybuf);
}

LOG(fmt, a1, a2, a3, a4, a5)
char *fmt;
{
	if (logfile != NULL)
	{
		fprintf(logfile,fmt,a1,a2,a3,a4,a5);
		fflush(logfile);
	}
}


/****************************************************************************
 *
 *	isbinfile - return true if binary file, else false.
 *
 */

isbinfile(fn)
FILE *fn;
{
	register int c, i;

	for(i=0; i<1000; i++) {
		if ((c = fgetc(fn)) == EOF) break;
		if(c==0 || c>127) {
			image = TRUE;
			if (debug) LOG("File is Binary.\n");
			return(1);
			}
		}
	image = FALSE;
	if (debug) LOG("File is Text.\n");
	return(0);
}

